contribute_url = 'https://github.com/SchmollerLab/Cell_ACDC/blob/main/CONTRIBUTING.rst'

github_url = 'https://github.com/SchmollerLab/Cell_ACDC'

issues_url = 'https://github.com/SchmollerLab/Cell_ACDC/issues'

forum_url = 'https://github.com/SchmollerLab/Cell_ACDC/discussions'

resources_url = 'https://github.com/SchmollerLab/Cell_ACDC#resources'

my_contact_url = 'https://www.helmholtz-munich.de/ife/about-us/people/staff-detail/ma/8873/Dr.-Padovani/index.html'

user_manual_url = 'https://github.com/SchmollerLab/Cell_ACDC/blob/main/UserManual/Cell-ACDC_User_Manual.pdf'

cite_url = 'https://bmcbiol.biomedcentral.com/articles/10.1186/s12915-022-01372-6'